/* optimized multiplication routine */

#include "optmult.h"

/*
   rationale:

      The other multiplication routines are optimized for the task
      of computing the product  ab  for fixed Milnor basis elements
      a  and  b.  In practice, however, one encouters the computation 
      of

               a ( b1 + b2 + b3 + ... + bN ),  N large

      and the optimization for this task is different.

      The routine

         optimized_multiply
            (stop *s, stoplist *diff, unit *(dat[]), modinfo *mi, int deg);

      computes  s * diff  and adds the result to the vectors dat[k] != NULL.

   plan:

      1)    
         Compute max(b1,b2,...,bN)  (taken entrywise)
         This gives restrictions for the entries in the admissible
         matrices.

         One day the maximum should be computed 
            a)  for each gen that is read from the datafile
            b)  for each gen that is newly introduced
         ie. max should be part of the internal representation of
         the resolution.

      2)    
         Compute according to the following scheme:

            for (admissible matrices for a) 
               go through all b's and compute the result
               add this result to dat.

*/

typedef struct {
  unit **vec;             /* pointers to vectors for the result */
  modinfo *mi;            /* module info, needed to access vec  */
  stoplist *diff;         /* the second factor                  */
  stop *first;
  int deg;
  element *basis;         /* for make_matrix_same_sig          */
  stop *sig;              /* fixed signature case              */
  unit *vecbas;           /* vector for the basis based computation */
} optmatrix_type;

optmatrix_type OM;

void make_max_stop(stop *res, stoplist *diff) {
  /* stop *s = diff->dat; int n = diff->num; */ 
   *((long long *) &(res->dat[0])) = 0xffffffffffffffffLL;
	/*
  clear_stop(res);
  while (n--) {
    int i;
    for (i=NALG;i--;) {
      int x = s->dat[i];
      if (x > res->dat[i]) res->dat[i] = x; 
    }
    s++;
  }
	*/
}

void init_optmatrix(stop *first, stop *max, unit *(dat[]), modinfo *mi, 
               stoplist *diff, int deg) {
  OM.vec  = dat;
  OM.mi   = mi;
  OM.diff = diff;
  OM.first = first;
  OM.deg = deg;
}

#define TOCP(x) ((unsigned char *) x)

inline
int firstpart(int n, unit msk, unit *dat) {
  unsigned c = NALG, deg = OM.first->dat[n-1];
  while (c--) {
	 unsigned x = deg >> c;
	 if (c) {
		unsigned badbits = TOCP(&msk)[(c-1)];
		REMOVE_BADBITS(x,badbits);
	 }
	 TOCP(dat)[c] = x;
	 deg -= x << c;
  }
  /* if (0 != deg) return 0; */ /* always returns true */
  return 1;
}

#if 1 || defined USEMMX
inline /* assembler part contains labels, so must not be inlined */
#endif
int nextpart(int n, unit msk, unit *dat) {
  unsigned c,deg=TOCP(dat)[0],x;
#if 0 && defined USEMMX /* Pentium optimization looks different */
  asm volatile ("movl %0,%1\n\tmovq (%1),%%mm0\n\t" 
					 : : "m" (dat), "r" (n) : "%1");  /* now mm0 = *dat */
  asm volatile("movb $1,%%bl\n\t"
					"psrlq $8,%%mm0\n\t"
					"movd %%mm0,%%eax\n\t"
					"orl %%eax,%%eax\n\t"
					"jnz .kl1\n\t"
					"orb $4,%%bl\n\t"
					"psrlq $32,%%mm0\n\t"
					"movd %%mm0,%%eax\n\t"
					".align 16\n.kl1:\n\t"
					"orw %%ax,%%ax\n\t"
					"jnz .kl2\n\t"
					"orb $2,%%bl\n\t"
					"shrl $16,%%eax\n\t"
					".align 16\n.kl2:\n\t"
					"orb %%al,%%al\n\t"
					"jnz .kl3\n\t"
					"incb %%bl\n\t"
					"orb %%ah,%%ah\n\t"
					"jnz .kl3\n\t"
					"incb %%bl\n\t"
					".align 16\n.kl3:\n\t"
					"movb %%bl,%0\n\t"
					/*"emms\n\t"*/
					: "=r" (c) : : "%eax", "%bl"
					);
#else  
  c=1; while ((c < NALG) && (0 == TOCP(dat)[c])) c++; 
#endif  
  
  if (c >= NALG) return 0;
  x = TOCP(dat)[c] - 1;
  {
	 unsigned badbits = TOCP(&msk)[c-1];
	 REMOVE_BADBITS(x,badbits);
  }
  deg += (TOCP(dat)[c] - x) << c;
  TOCP(dat)[c] = x; 
  {
	 while (--c) { /* c is at most 6 */
		x = deg >> c;
		{
		  unsigned  badbits = TOCP(&msk)[c-1];
		  REMOVE_BADBITS(x,badbits);
		} 
		TOCP(dat)[c] = x;
		deg -= x << c;
	 }
  }
  /* now c == 0 */
  TOCP(dat)[0] = deg;
  return 1;
}

long long glaux = 0x8080808080808080LL;

inline
void initgetloop(unit *sum, unit *msk) {
  asm("movq (%0),%%mm6\n\t"     /* mm6 = *sum */
		"movq (%1),%%mm7\n\t"     /* mm7 = *msk */
		"movq glaux,%%mm5\n\t"    /* mm5 = glaux */
		"psubb %%mm5,%%mm6\n\t"   /* mm6 -= mm5  */
		: /* no output */
      : "r" (sum), "r" (msk)
		);
}

#define MMXGETRES(s,res,LB) { \
	asm volatile("movq (%1),%%mm1\n\t" \
   "pxor %%mm5,%%mm1\n\t"\
	"movq %%mm6,%%mm0\n\t"\
	"pcmpgtb %%mm1,%%mm0\n\t"\
	"packsswb %%mm0,%%mm0\n\t"\
	"movd %%mm0,%0\n\t"\
	"orl %0,%0\n\t"\
	"jnz .m" LB "0\n\t"\
	"psubb %%mm6,%%mm1\n\t"\
	"movq %%mm7,%%mm0\n\t"\
	"pand %%mm1,%%mm0\n\t"\
	"packsswb %%mm0,%%mm0\n\t"\
	"movd %%mm0,%0\n\t"\
	"orl %0,%0\n\t"\
	"jnz .m" LB "0\n\t"\
	"por %%mm7,%%mm1\n\t"\
	"movl $-1,%0\n\t"\
	"movq %%mm1,(%2)\n\t"\
	"jmp .m" LB "1\n\t"\
	".align 16\n"\
".m" LB "0:\n\t"\
	"xorl %0,%0\n\t"\
	".align 16\n"\
".m" LB "1:	  \n\t"\
: "=r" (ok) : "0" (s), "r" (res));\
   }

#define MMXGETRESII(s,res,LB) { \
	asm volatile ("movq (%1),%%mm1\n\t" \
   "pxor %%mm5,%%mm1\n\t"\
	"movq %%mm6,%%mm3\n\t"\
	"pcmpgtb %%mm1,%%mm3\n\t"\
	"psubb %%mm6,%%mm1\n\t"\
	"movq %%mm7,%%mm0\n\t"\
	"pand %%mm1,%%mm0\n\t"\
	"packsswb %%mm0,%%mm0\n\t"\
	"movd %%mm0,%0\n\t"\
	"orl %0,%0\n\t"\
	"jnz .m" LB "0\n\t"\
	"packsswb %%mm3,%%mm3\n\t"\
	"movd %%mm3,%0\n\t"\
	"orl %0,%0\n\t"\
	"jnz .m" LB "0\n\t"\
	"por %%mm7,%%mm1\n\t"\
	"movl $-1,%0\n\t"\
	"movq %%mm1,(%2)\n\t"\
	"jmp .m" LB "1\n\t"\
	".align 16\n"\
".m" LB "0:\n\t"\
	"xorl %0,%0\n\t"\
	".align 16\n"\
".m" LB "1:	  \n\t"\
: "=r" (ok) : "0" (s), "r" (res));\
   }


/* 
	It might be good to insert
     "add $20,%1\n\t"\
     "prefetch (%1)\n\t"\
	in the code below.  Unfortunately, gas doesnt know about AMDs prefetch
   instruction.
     "prefetch (%eax)" would be  0x0f 0x0d 0x00
*/

#define MMXGETRESIII(s,res,LB) { \
	asm volatile ("movq (%1),%%mm1\n\t" \
	"movq %%mm6,%%mm3\n\t"\
   "pxor %%mm5,%%mm1\n\t"\
	"pcmpgtb %%mm1,%%mm3\n\t"\
	"psubb %%mm6,%%mm1\n\t"\
	"movq %%mm7,%%mm0\n\t"\
	"pand %%mm1,%%mm0\n\t"\
	"packsswb %%mm0,%%mm0\n\t"\
	"movd %%mm0,%0\n\t"\
	"orl %0,%0\n\t"\
	"jnz .m" LB "1\n\t"\
	"packsswb %%mm3,%%mm3\n\t"\
	"movd %%mm3,%0\n\t"\
	"orl %0,%0\n\t"\
	"jnz .m" LB "1\n\t"\
	"por %%mm7,%%mm1\n\t"\
	"movq %%mm1,(%2)\n\t"\
	".align 16\n"\
".m" LB "1:	  \n\t"\
: "=r" (ok) : "0" (s), "r" (res));\
   }


/*
inline char getresult(stop *s, stop *res) {  				 
  unsigned char c,nok;
  for (nok=c=0;c < NALG;c++) {
	 short int x = (s->dat[c]);
	 if (nok |= (x < OM.sum[0][c+1])) break;
	 x -= OM.sum[0][c+1];
	 if (nok |= (x & OM.msk[0][c+1])) break;
	 res->dat[c] = x | OM.msk[0][c+1];
  }
  return !nok;
}
*/

void ohr(int n, unit sum, unit msk);

inline
void opt_handle_row(int n) {
  ohr(n,0,0);
}

void pt(int n) {
  n = 10-n; while (n--) printf(" ");
}

void ohr(int n, unit sum, unit msk) {
  unit newsum, newmsk, dat;
  if (n) {
	 if (firstpart(n,msk,&dat)) {
		do {
		  /*pt(n); printf(" msk=%Lx dat=%Lx\n",msk,dat);*/
		  #ifdef USEMMX
		  asm volatile ("leal %0,%1\n\tmovq (%1),%%mm0\n\t" : : "m" (dat) , "r" (n) : "%1");
		  asm volatile ("movq %0,%%mm1\n\t" : : "m" (msk));
		  asm volatile ("psllq $8,%mm1\n\t");
		  asm volatile ("movq %0,%%mm2\n\t" : : "m" (sum));
		  asm volatile ("por %mm0,%mm1\n\t");
		  asm volatile ("psrlq $8,%mm0\n\t");
		  asm volatile ("paddb %mm0,%mm2\n\t");
		  asm volatile ("movq %%mm1,%0\n\t" : : "m" (newmsk));
		  asm volatile ("movq %%mm2,%0\n\t" : : "m" (newsum));
		  #else
		  newmsk = (msk << 8) | dat;
		  newsum = sum + (dat >> 8);
		  #endif
		  ohr(n-1,newsum,newmsk);
		} while (nextpart(n,msk,&dat));
	 }
	 return;
  } else {   /* get result */
	 stop *s = OM.diff->dat; int n = OM.diff->num;
	 /* initgetloop(&TOU(OM.sum[0][1]), &TOU(OM.msk[0][1])); */
	 initgetloop(&sum, &msk);
	 while (n--) {
		int ok;
		stop res;
		MMXGETRESIII(s,&res,"X");
		if (!ok) {
		  int k; 
		  unsigned int z; int **aux = (int **) OM.vec;
		  res.id = s->id;
		  k = module_seqno_with_degree(&res,OM.mi,OM.deg);
		  z = 1 << (k & (UNMASK >> 1)); k >>= UNSIZE-1;
		  while (NULL != *aux) {
			 (*aux)[k] ^= z;
			 aux++;
		  }
		}
		s++;
	 }
  }
}

void optimized_multiply(stop *s, stoplist *diff, unit *(dat[]), 
								modinfo *mi, int deg) {
  int c;
  stop max;
  make_max_stop(&max,diff);
  init_optmatrix(s,&max,dat,mi,diff,deg);
  for (c=NALG;c--;) 
	 if (s->dat[c]) {
		opt_handle_row(c+1);
		MMXCLEAR;
		return;
	 }
  opt_handle_row(0);
  MMXCLEAR; /* clear mmx status */	
}

/**************************************************************************

   same optimization for make_matrix_same_sig

**************************************************************************/

void init_optmatrix_basis(stop *first, stop *max, unit *dat, element *basis, 
               stoplist *diff, stop *sig) {
  OM.vecbas = dat;
  OM.basis  = basis;
  OM.diff   = diff;
  OM.first  = first;
  OM.sig    = sig;
}

void ohrb(int n, unit sum, unit msk);

inline
void opt_handle_row_basis(int n) {
  ohrb(n,0,0);
}

stop OMBauxstop;

void ohrb(int n, unit sum, unit msk) {
  unit newsum, newmsk, dat;
  if (n) {
	 if (firstpart(n,msk,&dat)) {
		do {
		  #ifdef USEMMX
		  asm volatile ("leal %0,%1\n\tmovq (%1),%%mm0\n\t" : : "m" (dat) , "r" (n) : "%1");
		  asm volatile ("movq %0,%%mm1\n\t" : : "m" (msk));
		  asm volatile ("psllq $8,%mm1\n\t");
		  asm volatile ("movq %0,%%mm2\n\t" : : "m" (sum));
		  asm volatile ("por %mm0,%mm1\n\t");
		  asm volatile ("psrlq $8,%mm0\n\t");
		  asm volatile ("paddb %mm0,%mm2\n\t");
		  asm volatile ("movq %%mm1,%0\n\t" : : "m" (newmsk));
		  asm volatile ("movq %%mm2,%0\n\t" : : "m" (newsum));
		  #else
		  newmsk = (msk << 8) | dat;
		  newsum = sum + (dat >> 8);
		  #endif
		  ohrb(n-1,newsum,newmsk);
		} while (nextpart(n,msk,&dat));
	 }
	 return;
  } else {   /* get result */
	 stop *s = OM.diff->dat; int n = OM.diff->num;
	 initgetloop(&sum,&msk);
	 for (;n--;s++) {
		int ok;
		stop res;
		MMXGETRESIII(s,&res,"Y");
		if (!ok) {
		  int k; int z;
		  res.id = s->id;       
		  set_signature(&res,OM.sig);
		  /*fprintf_stop(stdout,&res);*/
		  k = num_stop_same_signature(OM.basis,&res);
		  if (-1 == k) {
			 printf("\nstop "); fprintf_stop(stdout,&res);
			 printf(" not found\n"); 
			 printf("signature = ");
			 fprintf_stop(stdout,OM.sig);
			 printf("\n");
			 printf("basis = ");
			 print_element(OM.basis);
			 error("opt_make_matrix_same_sig");
		  } else {
			 z = 1 << (k & (UNMASK >> 1)); 
			 ((int *) (OM.vecbas))[k >> (UNSIZE - 1)] ^= z;
		  }
		}
	 }
  }
}

void optimized_multiply_basis(stop *s, stoplist *diff, unit *dat, 
                     element *basis, stop *sig) {
  stop max,aux; int n; 
/*
printf("opt_mult_basis: ");
fprintf_stop(stdout,s);
printf(" * ");
print_stoplist(diff);
printf(" = ");
*/
  copy_stop(&aux,s);
  for (n=0;n<NALG;n++) aux.dat[n] |= theprofile.mask[n];
  make_max_stop(&max,diff);
  init_optmatrix_basis(&aux,&max,dat,basis,diff,sig);
  for (n=NALG;n--;) 
	 if (aux.dat[n]) {
		opt_handle_row_basis(n+1);
		return;
	 }
  opt_handle_row_basis(0); 
}

void opt_make_matrix_same_sig(
	module *ms, module *mt, int t, signature *sig,
   small_element *bs, small_element *bt, matrix *mat) {
  int n;
  stoplist *diff; int numgen;
  /* first make bases */
  
  for (n=NALG;n--;) OMBauxstop.dat[n] = theprofile.mask[n];
  
  modulebasis((element *) bs,ms,t,sig);
  modulebasis((element *) bt,mt,t,sig);
  
  set_values(mat,bs->num,bt->num);
  clear_matrix(mat,bs->num,bt->num);
  for (n=0;n<bs->num;n++) {
	 numgen = bs->dat[n].id; diff = &(ms->gens[numgen].diff);
	 
	 optimized_multiply_basis(&(bs->dat[n]),diff,mat->dat[n],(element *) bt,sig);
   
  }

  MMXCLEAR; /* clear MMX status */
}


