/* resolve.c   -     main program, computes Ext^{s,t} for fixed s */

/* #define MOORE */  

#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include "linalg.h"
#include "stoputils.h"
#include "modules.h"
#include "error.h"
#include "seqno.h"
#include "optmult.h"
#include "random.h"
#include "options.h"

#define ELP (element *)

/* some routines that probably belong in a different place, but so what */

void set_max_profile_automatic(module *m, int s, int t) {
  int lowdim, updim;
  if (t >= 15*s) { /* low profile probably better */
	 set_max_profile_lower(s,t);
	 lowdim = countmodulebasis(m,t,2*MAXELEMENT);
	 set_max_profile_upper(s,t);
	 updim = countmodulebasis(m,t,lowdim);
  } else { /* up profile probably better */
	 set_max_profile_upper(s,t);
	 updim = countmodulebasis(m,t,2*MAXELEMENT);
	 set_max_profile_lower(s,t);
	 lowdim = countmodulebasis(m,t,updim);
  }
  if ((lowdim >= MAXELEMENT) && (updim >= MAXELEMENT)) {
	 printf("\nlow profile dim >= %d\n up profile dim >= %d\n",lowdim,updim);
	 error("MAXELEMENT too small\n");
  }
  if (lowdim <= updim) { /* take low profile */
	 set_max_profile_lower(s,t);
  } else { /* take up profile */
	 set_max_profile_upper(s,t);
  }
}

void set_max_profile_high(int s, int t) {
  if ((t >= 7*s)) {
    set_max_profile_lower(s,t); return;
  }
  if (t-s <= 1) {
    set_empty_profile(); return;
  }
  set_max_profile_upper(s,t);
}

void show_difftime(FILE *fp, time_t t1, time_t t0) {  /* t1 = start time */
  int d,h,m,s;
  unsigned long diff = (unsigned long) difftime(t1,t0);
  s = diff % 60; diff /= 60;
  m = diff % 60; diff /= 60;
  h = diff % 24; diff /= 24;
  d = diff;
  if (d) {
	 fprintf(fp,"%dd %dh %dm",d,h,m);
  } else {
	 if (h) {
		fprintf(fp,"%dh %dm %ds",h,m,s);
	 } else {
		if (m) {
		  fprintf(fp,"%dm %ds",m,s);
		} else 
		  fprintf(fp,"%ds",s);
	 }
  }
}

void info_module(FILE *fp, module *m) {
  int k;
  fprintf(fp,"C%d: %d gens of degrees ",m->sdeg,m->numgens);
  for (k=0;k<m->numgens;k++) fprintf(fp," %d",m->gens[k].deg);
/*   
   printf("\n");
   printf("Differentials:");
   for (k=0;k<m->numgens;k++) {
      printf(" %d --> ",k);
      print_stoplist(m->gens[k].diff);
   }
*/
}

void printpercentage(float f) {
  int i = (int) (10 * f);
  int a = i / 10, b = i % 10;
  cursor_back(6);
  printf("%3d.%1d%%",a,b);
  fflush(stdout);
}

int numsum, numunion, numintersection;
float diffpercentage;

void diff_statistic(int num, unit *(dt[]), modinfo *mi, int numunits) {
  /* Count the number of bits in the intersection (resp. union) of the */
  /* new dg's.  Estimate the need for an optimized computation of ddg  */
  int n,m; 
  if (num <= 1) return;
  numsum = numunion = numintersection = 0;
  for (m=0;m<numunits;m++) {
	 unit xu,xi;
	 xi = xu = 0; xi = ~xi;
	 for (n=0;n<num;n++) {
		numsum += alpha(dt[n][m]);
		xi &= dt[n][m]; xu |= dt[n][m];
	 }
	 numunion += alpha(xu); numintersection += alpha(xi);
  }
  diffpercentage = numunion; 
  diffpercentage /= numsum;
  diffpercentage *= 100;
}

int thes,thet;

void AHSS_optimize(int num, unit *(dt[]), modinfo *mi, int numunits) { 
  /* int num_gens,unit **dt2,modinfo *mi2 */
  unit u; int h,n,m,k,dm,numold;
  if (num == 1) return;
  dm = mi->totdim; numold=num;
  u = 1; u <<= ((dm-1) & UNMASK); h = (dm-1) >> UNSIZE;
  while (num > 0) {
	 if (0 == dm) error("dm=0 in AHSS_optimize()");
	 for (n=0;(n<num) && (0 == (dt[n][h] & u));n++) ;
	 if (n<num) {  /* found something */
		for (m=n+1;m<numold;m++) {  /* was num */
		  if (0 != (dt[m][h] & u)) { /* reduce */
			 for (k=numunits;k--;) 
				dt[m][k] ^= dt[n][k];
		  }
		}
		/* now move n-th entry to the end */
		{ 
		  unit *aux;
		  aux = dt[n]; dt[n] = dt[num-1];
		  dt[num-1] = aux; num--;
		}
	 }
	 dm--; u >>= 1; if (0 == u) { u = 0; u--; u >>= 1; u = ~u; h--; }
  }
}

void mysigint(int c) {
  printf("\ninterrupted by signal %d\n",c);
  printf("program termination should be safe\n");
  exit(1);
}

void setsignals(void) {
  signal(SIGINT,mysigint);
  signal(SIGQUIT,mysigint);
  signal(SIGTERM,mysigint);
}

void ignoresignals(void) {
  signal(SIGINT,SIG_IGN);
  signal(SIGQUIT,SIG_IGN);
  signal(SIGTERM,SIG_IGN);
}

#define NOQ if (!op_quiet)  /* use NOQ in front of printf(..) */
#define PER if (op_per)     /* use this for percentage info   */

int main(int argc, char **argv) {
  int s,t,tmax,complete,maxdegree;
  module mprev,mcurr,mnext;
  matrix *matp1,*matp2;
  minimatrix *minip1,*minip2; 
  signature sig;
  small_element bprev,bcurr,bnext; 
  FILE *fp=NULL; char fnme[13];
  FILE *flog=NULL; time_t tstart;
  char logfilename[50];

  setsignals();

  parse_options(argc,argv);

  printf("resolve.c -- looking for new generators in s = %d, t <= %d\n",arg_s,arg_t);

  matp1 = malloc(sizeof(matrix));
  matp2 = malloc(sizeof(matrix));
  minip1 = malloc(sizeof(minimatrix));
  minip2 = malloc(sizeof(minimatrix));

  if ((NULL == matp1) || (NULL == matp2) 
      || (NULL == minip1)) {
    printf("not enough memory for matrices\n");
    printf("tried to allocate %d bytes\n",2*sizeof(matrix) + 2*sizeof(minimatrix));
    error("change MAXDIM");
  }

  if (op_pc) op_quiet = 1;
  if (op_quiet) op_per = 0;
  if (op_test) op_recomp = 1;
  
  s = arg_s; tmax = arg_t;
  
  if (s <= 0) error("s should be >= 1");
  
  initialize_stops();
  set_empty_profile();
  if (-1 == op_alg) {
    set_algebra_Ainfty();
  } else {
    set_algebra_An(op_alg);
    printf("resolving %s\n",algebra_name(&thealgebra));
  }

  if (*logfileextension) {
    sprintf(logfilename,"%s/log.%s",algebra_path_name(&thealgebra),
	    logfileextension);
  } else {
    sprintf(logfilename,"%s/log",algebra_path_name(&thealgebra));
  }
  flog = fopen(logfilename,"at");
  if (NULL == flog) error("couldn't open log file");
  
  init_stoplist_buffer();
  
  if (s >= 2) {
    if (s >= 3) {
      read_module_no_diffs(&mprev,s-2);
    } else make_C0_module(&mprev);
    read_module(&mcurr,s-1);
    read_module(&mnext,s);
  } else {  /* s = 1 */
    read_module(&mnext,1);
    make_C0_module(&mcurr);
    make_zero_module(&mprev); mprev.sdeg = -1;
  }
   
  sprintf(fnme,"%s/C%d",algebra_path_name(&thealgebra),s);
  complete = mnext.header.data.tmax;    /* degrees that have already been done */
  maxdegree = mcurr.header.data.tmax+1; /* degrees that can be done            */

/*
info_module(&mprev);
info_module(&mcurr);
info_module(&mnext);
*/
  s--; 

  if (op_recomp) {
    t = s;            /* recompute known bidegrees */
  } else {
    t = complete+1; 
  }

#if defined MOORE
  if (t<s+2) t = s+2;       /* t = s+2 for the mod 2 Moore space */
#else
  if (t<s+1) t = s+1;       /* t = s+1 for the sphere            */
#endif

  if (op_test && (tmax>complete)) tmax = complete;

  if (tmax>maxdegree) {
    if (op_test) tmax = maxdegree; 
    NOQ printf("degree %d requested, only %d possible\n",arg_t,tmax);
  }

  for (;t<=tmax;t++) {
      
    if (t>maxdegree) {
      printf("degree %d not possible yet.\n",t);
      printf("restart when previous module present.\n");
      exit(2);
    }

    NOQ printf("(%d,%d)",s+1,t);

    tstart = time(NULL);

    thes = s+1; thet = t;

    set_max_profile_automatic(&mcurr,s,t);
    NOQ fppf(stdout,&theprofile);
         
    /* sort_profile(); */

    clear_stop(&sig);

    opt_make_matrix_same_sig(&mcurr,&mprev,t,&sig,&bcurr,&bprev,matp1);
    NOQ printf(" %dx%d ",matp1->num_r,matp1->num_c);
    NOQ { printf(" >K"); fflush(stdout); };
    kernel_basis(matp1,matp2);
    NOQ { printf("<"); fflush(stdout); };
    opt_make_matrix_same_sig(&mnext,&mcurr,t,&sig,&bnext,&bcurr,matp1);
    NOQ printf(" %dx%d ",matp1->num_r,matp1->num_c);
    NOQ { printf(" >Q"); fflush(stdout); };
    quotient_basis(matp2,matp1,minip1);
    NOQ { printf("<"); fflush(stdout); };

#define REQUNITS 3000  /* 2000 x 1000 bits per vector */
#define MAXUNITS (400 * MINIDIM * (REQUNITS >> UNSIZE))

    if (minip1->num_r > 0) { /* new generators needed */
      int k,l,num_gens=minip1->num_r;
      unit dat[MAXUNITS];
      unit *auxunitpointer;
      int n; float f; 
      int numsteps = 1, maxsteps = 1 << theprofile.numpts;
      int nummatrixcomp=1; 
      int stepmask = (maxsteps-1) >> 9; 
      /* percentage info in maximal 2^9 = 512 blocks */  
      
      volatile clock_t matclock=0,diffclock=0,auxclock;
		/* volatile, so compiler won't optimize them away */
      
      int numbot = mnext.numgens; 
      /* we're about to introduce generators */
      /* numbot,...,numbot+num_gens-1        */
      stoplist *(realdiffs[MINIDIM]);
      unit *(dt[MINIDIM]);   /* pointers to error terms          */
      unit *(dt2[MINIDIM]);  /* pointers to vectors for new gens */
      int num_units;   /* units per error term */
      int num_units2;  /* units per dg         */
      modinfo mi;   /* module_info for error terms */
      modinfo mi2;  /* module_info for dg          */
      
      NOQ printf(" %d new gen%c\n",num_gens, (num_gens > 1) ? 's' : ' ');
      if (MINIDIM < num_gens) error("MINIDIM too small");
    
      if (t <= complete) {
		  printf("\nModule claimed to be complete through degree %d\n",complete);
		  error("work this out!");
      }

      for (numbot=mnext.numgens,l=0;l<num_gens;l++) {
		  mnext.gens[numbot+l].deg = t;
		  mnext.gens[numbot+l].num = numbot+l;
		  realdiffs[l] = &(mnext.gens[numbot+l].diff);
		  realdiffs[l]->num = 0; /* really needed ? */
      }

      /* now new empty generators have been declared */
      
      make_modinfo(&mi,&mprev,t); 
      make_modinfo(&mi2,&mcurr,t); 
      
      num_units = 1 + (mi.totdim >> UNSIZE); 
      /* number of units needed for each error term */
      num_units2 = 1 + (mi2.totdim >> UNSIZE); 
      /* number of units needed for each vector */
      
      auxunitpointer = dat;
      for (l=0;l<num_gens;l++,auxunitpointer+=num_units) 
		  dt[l] = auxunitpointer;
      for (l=0;l<num_gens;l++,auxunitpointer+=num_units2) 
		  dt2[l] = auxunitpointer;
      
      if ((auxunitpointer - dat) > MAXUNITS) {
		  printf(" dim = %d ",mi.totdim);
		  printf("\nunits needed: (%d + %d) times %d \n",num_units,num_units2,num_gens);
		  printf("units available: %d\n",MAXUNITS);
		  error("not enough space, change MAXUNITS");
      }
      
      for (l=0;l<num_gens;l++) 
		  for (n=num_units2;n--;) dt2[l][n] = 0; /* diff vector cleared */
      
      for (k=0;k<num_gens;k++) 
		  add_basis_vector_to_module_vector(minip1->dat[k],ELP &bcurr,dt2[k],&mi2);
	
      NOQ printf(" dim = %d ",mi.totdim);
		NOQ fflush(stdout);

      for (l=0;l<num_gens;l++) 
		  for (n=num_units;n--;) dt[l][n] = 0; /* error vector cleared */
         
      /* compute initial error terms */
      auxclock = clock();
      add_diff(dt,num_gens,minip1,ELP &bcurr,&mcurr,&mi,t); 
      diffclock += (clock() - auxclock);
      
      PER printf("      ");

      clear_stop(&sig);
	
      while (next_signature(&sig)) {
		  /* find correction terms for this signature */

		  PER if (0 == (numsteps & stepmask)) {
			 f = numsteps; f *= 100; f /= maxsteps;
			 printpercentage(f);
		  }
		  numsteps++;
		  
		  modulebasis(ELP &bprev,&mprev,t,&sig);
		  clear_matrix((matrix *) minip1, num_gens, bprev.num);
		  for (l=0;l<num_gens;l++) 
			 extract_diff(minip1->dat[l],ELP &bprev, dt[l], &mi);
		  
		  /* now: mini1.dat[...] = error terms of this signature */
		  
		  if (!matrix_is_zero((matrix *) minip1)) {
			 auxclock = clock();
			 opt_make_matrix_same_sig(&mcurr,&mprev,t,&sig,&bcurr,&bprev,matp1);
			 
			 nummatrixcomp++;
			 
			 if (!lift(minip1,matp1,minip2)) {
				info_module(stdout,&mcurr);
				printf("Basis mcurr:");
				print_element(ELP &bcurr);
				error("Lift not found");
			 }; 
			 matclock += (clock() - auxclock);
			 
			 for (l=0;l<num_gens;l++) 
				add_basis_vector_to_module_vector(minip2->dat[l], ELP &bcurr,dt2[l],&mi2);
			 
			 /* correction terms have been added to  dt2[] */
			 
			 auxclock = clock();
			 add_diff(dt,num_gens,minip2,ELP &bcurr,&mcurr,&mi,t); 
			 diffclock += (clock() - auxclock);
			 /* error terms have been updated */                     
			 
		  }
		  
      } /* all signatures done */

      /* Now: gens are corrected. Check this is true */
		
      for (l=0;l<num_gens;l++)
		  for (n=0;n<num_units;n++)
			 if (0 != dt[l][n]) error("correction apparantly failed");

      fprintf(flog,"(%d,%d) ",s+1,t);
      fppf(flog,&theprofile); 
      /* profile will be reset by new_stoplist_from_vector */
		
      diff_statistic(num_gens,dt2,&mi2,num_units2);
		
      AHSS_optimize(num_gens,dt2,&mi2,num_units2);
		
      for (l=0;l<num_gens;l++) {
		  mnext.gens[numbot+l].num = 
			 new_stoplist_from_vector(&(mnext.gens[numbot+l].diff),
											  dt2[l],&mi2);
      }
		
      mnext.numgens += num_gens;   /* new gens officially declared */
		
      ignoresignals();    /* no user break while updating the resolution */
		
      fp = fopen(fnme,"ab");
		
      for (l=0;l<num_gens;l++) {
		  write_gen(fp,&(mnext.gens[numbot+l]));
		  fflush(fp);
      }
		
      fclose(fp);

      fprintf(flog," dim %d ",mi.totdim);
      fprintf(flog," %d gens  ",num_gens);
      show_difftime(flog,time(NULL),tstart);
      if (nummatrixcomp != maxsteps) 
		  fprintf(flog," (%d/%d mats) ",nummatrixcomp,maxsteps);
      fprintf(flog," (%ld mat, %ld diff) ",matclock,diffclock);
      if (num_gens > 1) 
		  fprintf(flog," (%d ->%3.1f%%) ",numunion,diffpercentage);
      fprintf(flog," summands: ");
      for (l=0;l<num_gens;l++) {
		  if (l) fprintf(flog,", ");
		  fprintf(flog,"%d",mnext.gens[numbot+l].num);
      }
      fprintf(flog,"\n");
      fflush(flog);
		
      setsignals();  /* restore signals */
		
      PER cursor_back(6);
      NOQ {
		  show_difftime(stdout,time(NULL),tstart);
		  if (nummatrixcomp != maxsteps) 
			 printf(" (%d/%d mats) ",nummatrixcomp,maxsteps);
		  printf(" (%ld mat, %ld diff) ",matclock,diffclock);
		  if (num_gens > 1) 
			 printf(" (%d ->%3.1f%%) ",numunion,diffpercentage);
		  printf(" summands: ");
		  for (l=0;l<num_gens;l++) {
			 if (l) printf(", ");
			 printf("%d",mnext.gens[numbot+l].num);
		  }
      }
      
    } 
      
    if (t > complete) {
      ignoresignals();
      complete = mnext.header.data.tmax = t;
      update_header(&(mnext.header));
      setsignals();
    }
      
    NOQ printf(" Ok.\n");
    
  }
    
  if (NULL != flog) fclose(flog);
    
  return 0;
}
